library(ggplot2)
library(ggpubr)

## Performance EA vs grid
##
## FC_b = FC_a + beta  if  correlated
## FC_b = FC_random    otherwise

rfc  <- function (n)
{
    rnorm(n)
}

generate.perfect.fc <- function(n, n.corr)
{
    n.corr.left  <- n.corr %/% 2
    n.corr.right <- n.corr - n.corr.left
    half  <- n - n.corr.left - n.corr.right
    a <- (1:n) -  half
    b <- c(a[1:n.corr.left], sample(half) - n.corr, a[(n-n.corr.right+1):n] )
    
    dt <- data.table(a = a,
                     b = b,
                     corr = c(rep(T, n.corr.left), rep(F, half), rep(T, n.corr.right)) )
    
    dt
}

generate.perfect.better.fc <- function(n, n.corr)
{
    half <- n / 2
    dt <- data.table(a = (1:n) - half - 1)
    dt <- dt[order(-abs(a))]
    dt[, corr := c(rep(T, n.corr), rep(F, n - n.corr) ) ]

    n.middle <- n - n.corr
    n.range <- n.middle
    dt[, b := ifelse(corr, a, sample(n.range, n.middle) -  n.range / 2 - 1) ]
    
    dt
}

generate.fc <- function(n, n.corr)
{
    dt <- data.table(a = rfc(n) )
    dt <- dt[order(-abs(a))]
    dt[, corr := c(rep(T, n.corr), rep(F, n - n.corr) ) ]
    
    dt[, b := ifelse(corr, sign(a) * rexp(n, 1 / abs(a)), rfc(n))]
    
    dt
}

generate.better.fc <- function(n, n.corr){
    half  <- n / 2
    .sd <- (n - n.corr) / 2 / 3 ## / 3 ~= 0.999 quartile
    
    dt <- data.table(a = ( (1:n) - half - 1 ))
    dt <- dt[order(-abs(a))]
    dt[, corr := c(rep(T, n.corr), rep(F, n - n.corr) ) ]
    dt[, a := ifelse(corr, a, rnorm(n, 0, .sd)) ]
       
    ## brand <- rfc(n)
    ## brand <- brand[order(-abs(brand))]
    ## brand[(n.corr+1):n] <- brand[(n.corr+1):n][sample(n - n.corr)]
    
    dt[, b := ifelse(corr, sign(a) * rpois(n, abs(a)), rnorm(n, 0, .sd) ) ]
    
    dt
}


generate.down.fc <- function(n, n.corr, range = 100, n.noise = 0, randomize = TRUE)
{
    
    a <- sample(n.corr + n.noise, n.corr)
    b <- if (randomize) sample(n.corr + n.noise, n.corr) else a[1:n.corr]
    dt <- data.table(a = a, b = b)
    dt[, corr := rep(T, n.corr) ]
    dt
}

## generate.down.fc(13000, 5, range = 2, n.noise = 10)[]

generate.noisy.fc <- function(n, n.corr, range = 100, n.noise = 0, randomize = TRUE, two.quadrants = TRUE)
{
    half  <- n %/% 2
    other.half <- n - half

    if (range > 0)
        n.corr <- sample(n.corr:(n.corr + range), 1)

    if (n.noise == "n.corr")
        n.noise <- n.corr
 
    
    dt1 <- generate.down.fc(half, n.corr, range, n.noise, randomize = randomize)
    dt1[, a := a - half - 1]
    dt1[, b := b - half - 1]

    if (two.quadrants)
    {
        dt2 <- generate.down.fc(other.half, n.corr, range, n.noise, randomize = randomize)
        dt2[, a := other.half - a]
        dt2[, b := other.half - b]
        dt <- rbind(dt1, dt2)
    } else
        dt  <- dt1

    dt.random <- data.table(a = setdiff(sample(1:n) - half - 1, dt$a),
                            b = setdiff(sample(1:n) - half - 1, dt$b),
                            corr = F)

    rbind(dt, dt.random)
}


quad.stats <- function (dt, algorithm="ea", mm=NULL, nn=NULL)
{
    n.corr <- sum(dt$corr)
    n.corr.dd <- sum( dt[a < 0 & b < 0,]$corr )
    n.corr.uu <- sum( dt[a > 0 & b > 0,]$corr )
    n <- nrow(dt)
    P <- n.corr
    N <- n - n.corr
    
    if ("grid" == algorithm)
        algorithm  <- "classic"

    if( algorithm %in% c("ea", "classic") )
    {
        rr <- RedRibbon(dt$a, dt$b, enrichment_mode = "hyper-two-tailed")
        if ( is.null(mm) || is.null(nn) )
            mm <- nn <- if ("classic" == algorithm) ceiling( sqrt(nrow(dt)) ) else NULL ## like for RRHO2
        quad <- quadrants(rr, whole=FALSE, m=mm, n=nn, algorithm = algorithm, permutation=FALSE)
        
        union.pos <- union(if (n.corr.uu > 0) quad$upup$positions else c(),
                           quad$downdown$positions)
        
        TP <- nrow(dt[union.pos][TRUE == corr,])
        FP <- nrow(dt[union.pos][FALSE == corr,])
        TN <- nrow(dt[-union.pos,][FALSE == corr,])
        FN <- nrow(dt[-union.pos][TRUE == corr,])
    } else if (algorithm == "rrho2")
    {
        dt[, gene := 1:n]
        RRHO_obj <-  RRHO2_initialize(dt[, c("gene", "a")], dt[, c("gene", "b")], labels = c("a", "b"), log10.ind=TRUE, method="fisher")

        union.gene <- union(RRHO_obj$genelist_dd$gene_list_overlap_dd,
                            if (n.corr.uu > 0) RRHO_obj$genelist_uu$gene_list_overlap_uu else c())

        TP <- nrow(dt[gene %in% union.gene & TRUE == corr,])
        FP <- nrow(dt[gene %in% union.gene & FALSE == corr,])
        TN <- nrow(dt[ ( ! gene %in% union.gene ) & FALSE == corr,])
        FN <- nrow(dt[ ( ! gene %in% union.gene ) & TRUE == corr,])
    } else if (algorithm == "rrho")
    {
        dt[, gene := 1:n]
        RRHO_obj <-  originalRRHO(as.data.frame(dt[, c("gene", "a")]), as.data.frame(dt[, c("gene", "b")]), labels = c("a", "b"))

        union.gene <- union(RRHO_obj$downdown$intersection,
                            if (n.corr.uu > 0) RRHO_obj$upup$intersection else c())

        TP <- nrow(dt[gene %in% union.gene & TRUE == corr,])
        FP <- nrow(dt[gene %in% union.gene & FALSE == corr,])
        TN <- nrow(dt[ ( ! gene %in% union.gene ) & FALSE == corr,])
        FN <- nrow(dt[ ( ! gene %in% union.gene ) & TRUE == corr,])
    } else
        stop("quad.stats(): wrong algorithm not in: ea, classic, rrho2 or rrho")
    

    
    TPR <- TP / P  ## sensivity
    FPR <- FP / N
    TNR  <- TN / N ## specificity
    FNR <- FN / P

    FDR  <- FP / (FP + TP)
    ACC <- (TP + TN) / (P + N)
    
    list( ## quadrants = quad,
         stats = list(P = P, N = N, TP = TP, FP = FP, TN = TN, FN = FN,
                      TPR = TPR, FPR = FPR, TNR = TNR, FNR = FNR,
                      FDR = FDR, ACC = ACC)
         )
}

stats.grid.ea <- function (n = 23000, n.corr = 5000, generate.fn=generate.dt, n.samples=2*96, mc.cores=48, algorithms=c("ea", "grid", "rrho2", "rrho"))
{
    names(algorithms)  <- algorithms
    stats.dt <- do.call(rbind, mclapply(1:n.samples,
                                        function(index)
                                        {
                                            dt <- generate.fn(n, n.corr)
                                            ##
                                            alg <- lapply(algorithms, function (algorithm) quad.stats(dt, algorithm=algorithm) )
                                            ## print(alg)
 
                                            stats <- lapply(algorithms,
                                                            function (algorithm)
                                                            {
                                                                stats.alg <- as.data.table( alg[[algorithm]]$stats )
                                                                ## print(stats.alg)
                                                                colnames(stats.alg) <- paste0(colnames(stats.alg), '.', algorithm)
                                                                return(stats.alg)
                                                            })
                                            ## print(stats)
 
                                            names(stats) <- NULL
                                            ## print(unlist(stats))
                                            do.call(cbind, c(stats, deparse.level = 0))
                                        },
                                        mc.cores=mc.cores))

    stats.dt
}


## test.position <- function (stats.dt, column="TPR")
## {
##     maxp <- max(stats.dt[[paste0(column, '.grid')]], stats.dt[[paste0(column, '.ea')]], stats.dt[[paste0(column, '.rrho2')]])
##     minp <- min(stats.dt[[paste0(column, '.grid')]], stats.dt[[paste0(column, '.ea')]], stats.dt[[paste0(column, '.rrho2')]])

##     maxp + 0.1 * (maxp - minp)
## }

gg.cmp <- function (stats.dt, column="TPR", algorithms=c("ea", "classic", "rrho2", "rrho"), x.angle = 10)
{
    title.size <- 14
    text.size <- 16
    
    gg.dt <- data.table(Algorithm=character(), x=double(), P = integer())
    if ( "ea" %in% algorithms)
        gg.dt <- rbind(gg.dt,
                       data.table(Algorithm="RedRibbon\nEvol. Algo.",
                                  x=stats.dt[[paste0(column, ".ea")]],
                                  P = stats.dt$P.ea))
    if ( "classic" %in% algorithms || "grid" %in% algorithms)
        gg.dt <- rbind(gg.dt, data.table(Algorithm="RedRibbon\nPlaisier et al.", x=stats.dt[[paste0(column, ".grid")]],
                                  P = stats.dt$P.grid))
    if ( "rrho2" %in% algorithms)
        gg.dt <- rbind(gg.dt, data.table(Algorithm="Cahill et al.", x=stats.dt[[paste0(column, ".rrho2")]],
                                  P = stats.dt$P.rrho2) )
    if ( "rrho" %in% algorithms)
        gg.dt <- rbind(gg.dt, data.table(Algorithm="Plaisier et al.", x=stats.dt[[paste0(column, ".rrho")]],
                                  P = stats.dt$P.rrho) )

    
    ## gg.dt$Algorithm <- reorder(gg.dt$Algorithm, gg.dt$Algorithm)
    ## print(gg.dt)
    maxp <- max(gg.dt$x)
    minp <- min(gg.dt$x)
    delta <- maxp - minp

    ## x.angle <- 14

    gg.dt[, Algorithm := factor(Algorithm, levels = c("Plaisier et al.", "Cahill et al.", "RedRibbon\nPlaisier et al.", "RedRibbon\nEvol. Algo.")) ]
    
    ggplot(gg.dt, aes(x=Algorithm, y=x, fill=Algorithm)) + 
        geom_violin(show.legend = FALSE) + ## stat_boxplot(geom = "errorbar", width = 0.5, lwd=1) +
        ## geom_boxplot(notch = TRUE,  outlier.shape = NA, ## outlier.size = 2,
        ##              color="black",lwd=1, alpha = 0.5, show.legend = FALSE) +
        geom_point(## mapping = aes(color = P),
                   size=2, position = position_jitterdodge(jitter.width = 1.5),
                   color = "black", alpha=0.2, show.legend = FALSE) +
        stat_compare_means(paired = TRUE, label.y = maxp + 0.1 * delta) +
        xlab(NULL) + ylab(NULL) +
        scale_y_continuous(breaks=pretty(c(minp, maxp), 5),
                           limits=c(minp-0.05*delta, maxp+0.1*delta), labels = function(x) formatC(x, width = 7, digits = 5)
                           ) +
        ggtitle(column) +
        theme(legend.position = "none",
              text = element_text(size=text.size),
              plot.title = element_text(size=title.size),
              axis.text.x = element_text(angle = x.angle, vjust = 1, hjust=1))
    ##colMeans(stats.dt)+ theme(plot.title = element_text(size=22))

}


stats.iterate.stepsize <- function(n = 23000, n.corr = 5000, generate.fn=generate.better.fc, n.samples = 2*96, mc.cores = 48, stepsizes = c(seq(250, 50, -50), 0))
{
    res <- do.call(rbind,
                   lapply(stepsizes,
                          function(stepsize)
                          {
                              cat(paste("stepsize =", stepsize, "\n") )
                              res.dt <- do.call(rbind,
                                                mclapply(1:n.samples,
                                                         function (index)
                                                         {
                                                             dt <- generate.fn(n, n.corr)
                                                             ## print(dt)
                                                             if( 0 == stepsize)
                                                             {
                                                                 grid.alg <- quad.stats(dt, algorithm="ea")
                                                             } else
                                                             {
                                                                 number <- n %/% stepsize
                                                                 grid.alg <- quad.stats(dt, algorithm="classic", mm = number, nn =  number)
                                                             }
                                                             ## print("HERE")
                                                             ## print(as.data.table(grid.alg$stats))
                                                             as.data.table(grid.alg$stats)
                                                         },
                                                         mc.cores = mc.cores))


                              res.dt[, stepsize := stepsize]
                              res.dt
                          }))

    res[, label := ifelse(stepsize == 0, "ea", "classic") ]
    as.data.table(res)
}


ggIterate <- function (stats.dt, column="TPR")
{
    .max <- max(stats.dt[[column]])
    .min <- min(stats.dt[[column]])
    .max <- .max + (.max - .min) * 0.1
    stats.dt[, stepsize.factor := as.factor(stepsize)]
    ggplot(stats.dt, aes(x = stepsize.factor, y = get(column) ) ) +
        geom_violin(show.legend = FALSE) +
        geom_jitter(height = 0,
            size=2, ## position = position_jitterdodge(jitter.width = 1.5),
            color = "black", alpha=0.2, show.legend = FALSE) +
        ylab("") + xlab("Stepsize") + 
        ## ylim(c(.min, .max)) +
        scale_y_continuous(labels = function(x) formatC(x, width=5, format = "g"))  +
        ggtitle(column) +
        geom_text(data = data.table(x = "0", y = .max, label = "ea"), mapping = aes(x = x, y = y, label = label) )
    
}
